ARCHCFLAGS	=	-DSYSVBFUNC -DSYSVSTR -DNOGETDTBLSIZ -DSYSVSIGNAL \
				-DNOWAIT3 -DNOUNIXDOM -DRSHCOMMAND=\"/usr/bin/rsh\"
ARCHDLIB	=	-lsocket -lnsl
ARCHDOBJ	=
ARCHLIB		=	-lsocket -lnsl
HASRANLIB	=	f
AR			=	ar
PVM_ARCH	=	UXPM
MAKE		=	make
