ARCHCFLAGS	=	-DSOCKLENISUINT -DSYSVSIGNAL \
				-DRSHCOMMAND=\"/usr/bin/rsh\" -DSOCKADHASLEN \
				-DNEEDSSELECTH \
				-g -q64 -qmaxmem=-1 -qarch=pwr3 -DHASSTDLIB
ARCHFFLAGS	=	-q64
ARCHDLIB	=	-lbsd
ARCHDOBJ	=
ARCHLIB		=
HASRANLIB	=	t
AR			=	ar -X64
LOPT		=	-g -q64
PVM_ARCH	=	AIX56K64
NEEDF2M4	=	t
MAKE		=	make
