ARCHCFLAGS	=	-DSYSVSIGNAL -DNOUNIXDOM -DSOCKADHASLEN \
				-DRSHCOMMAND=\"/usr/bin/rsh\" -DNEEDSSELECTH
ARCHDLIB	=
ARCHDOBJ	=
ARCHLIB		=
HASRANLIB	=	t
PVM_ARCH	=	SP2MPI
AR_FRONT	=	ar
AR_NODE		=	ar
AR			=	ar
CC			=	cc
NODECC		=	mpcc
NEEDF2M4	=	t
FORTRAN		=	f77
NODEFORT	=	mpxlf
MAKE		=	make
