ARCHCFLAGS	=	-DSOCKLENISUINT -DSYSVSIGNAL \
				-DRSHCOMMAND=\"/usr/bin/rsh\" -DSOCKADHASLEN \
				-DNEEDSSELECTH
ARCHDLIB	=	-lbsd
ARCHDOBJ	=
ARCHLIB		=
HASRANLIB	=	t
AR			=	ar
PVM_ARCH	=	AIX56K
NEEDF2M4	=	t
MAKE		=	make
