cd src\os2
make install
cd ..\..\tracer\os2
make install
cd ..\..\console\os2
make install
cd ..\..\pvmgs\os2
make install
cd ..\..\hoster\os2
make all
cd ..\..\examples\os2
make all
cd ..\..\gexamples\os2
make all
cd ..\..\libfpvm\os2
make install
cd ..\..\rm\os2
make install
cd ..\..\xep\os2
make all
cd ..\..
