ARCHCFLAGS	=	-DNEEDSFFS -DNOUNIXDOM -DNOWAIT3 -DSHAREDTMP \
				-DSTATISTICS -DNOGETDTBLSIZ -DREALIX -DMAXPATHLEN=1024
ARCHDLIB	=
ARCHDOBJ	=
ARCHLIB		=
HASRANLIB	=	f
AR			=	ar
PVM_ARCH	=	M88K
MAKE		=	make
