ARCHCFLAGS	=	-DSYSVSIGNAL -DRSHCOMMAND=\"/usr/bsd/rsh\" -DNOWAIT3 \
				-64 -woff 1174 -DHASSTDLIB -DCTIMEISTIMET
ARCHFFLAGS	=	-64
#ARCHDLIB	=	-lsun
ARCHDOBJ	=
#ARCHLIB	=	-lsun
AR			=	ar
HASRANLIB	=	f
PVM_ARCH	=	SGI64
MAKE		=	make
