cd src\os2
make clean
cd ..\..\console\os2
make clean
cd ..\..\pvmgs\os2
make clean
cd ..\..\examples\os2
make clean
cd ..\..\gexamples\os2
make clean
cd ..\..\hoster\os2
make clean
cd ..\..\libfpvm\os2
make clean
cd ..\..\rm\os2
make clean
cd ..\..\tracer\os2
make clean
cd ..\..\xep\os2
make clean
cd ..\..
