ARCHCFLAGS	=	-DSYSVBFUNC -DSYSVSTR -DNOGETDTBLSIZ \
				-DSYSVSIGNAL -DNOWAIT3 -DNOUNIXDOM \
				-DRSHCOMMAND=\"/bin/rsh\" \
				-DHASSTDLIB -DNOREXEC -DNOWAIT3 -UNOWAITPID \
				-DNOSTRCASE -USOCKADHASLEN
ARCHDOBJ	=
ARCHDLIB	=	-lnsl -lsocket
ARCHLIB		=	-lnsl -lsocket
HASRANLIB	=	f
AR			=	ar
PVM_ARCH	=	UWARE
MAKE		=	make
