ARCHCFLAGS	=	-DSYSVBFUNC -DSYSVSTR -DNOGETDTBLSIZ -DSYSVSIGNAL \
				-DNOWAIT3 -DNOUNIXDOM -DRSHCOMMAND=\"/usr/bin/rsh\"
ARCHDLIB	=	-lnsl -lsocket
ARCHDOBJ	=
ARCHLIB		=	-lnsl -lsocket
HASRANLIB	=	f
AR			=	ar
PVM_ARCH	=	X86SOL2
MAKE		=	make
