ARCHCFLAGS	=	-DSOCKADHASLEN -DNOREXEC -DRSHCOMMAND=\"/usr/bin/rsh\" \
				-DHASSTDLIB -DNEEDMENDIAN -DHASERRORVARS
ARCHDLIB	=
ARCHDOBJ	=
ARCHLIB		=	-lrpc
HASRANLIB	=	t
AR			=	ar
PVM_ARCH	=	BSD386
MAKE		=	gmake
