ARCHCFLAGS	=	-DSOCKLENISUINT -DSYSVSIGNAL -DNOUNIXDOM -D_BSD \
				-DRSHCOMMAND=\"/usr/bin/rsh\" -DSOCKADHASLEN \
				-DNEEDSSELECTH -DSHMEM
ARCHDLIB	=	-lbsd -lpthreads
ARCHDOBJ	=
ARCHLIB		=	-lbsd -lpthreads 
HASRANLIB	=	t
AR			=	ar
PVM_ARCH	=	RS6KMP
MAKE		=	make
