ARCHCFLAGS	=	-DSYSVSIGNAL -DRSHCOMMAND=\"/usr/bsd/rsh\" \
				-DNOUNIXDOM -64 -woff 1174 -DHASSTDLIB -DSHMEM \
				-DCTIMEISTIMET
ARCHFFLAGS	=	-64
#ARCHDLIB	=	-lsun
ARCHDOBJ	=
#ARCHLIB	=	-lsun
HASRANLIB	=	f
AR			=	ar
PVM_ARCH	=	SGIMP64
MAKE		=	make
