ARCHCFLAGS	=	-DSYSVSIGNAL -DNOWAIT3 -DNOWAITPID -DSYSVSTR \
				-DRSHCOMMAND=\"/usr/ucb/rcmd\" -DNOUNIXDOM
ARCHDLIB	=	-lsocket
ARCHDOBJ	=
ARCHLIB		=	-lsocket
HASRANLIB	=	f
AR			=	ar
PVM_ARCH	=	IPSC2
MAKE		=	make
