ARCHCFLAGS	=	-DSYSVSIGNAL -DNOWAIT3 -DRSHCOMMAND=\"/usr/bin/rsh\" \
				-DNEEDENDIAN -DFDSETNOTSTRUCT -DHASERRORVARS \
				-DCTIMEISTIMET -DSYSERRISCONST -DNOTMPNAM \
				-DNOUNIXDOM -DNOPROT
ARCHDLIB	=
ARCHDOBJ	=
ARCHLIB		=
HASRANLIB	=	t
PVM_ARCH	=	LINUX
AR_FRONT	=	ar
AR_NODE		=	ar
AR			=	ar
CC			=	cc
NODECC		=	cc
FORTRAN		=	f77
NODEFORT	=	f77
MAKE		=	make
