ARCHCFLAGS	=	-DNOWAIT3 -DNOUNIXDOM 
ARCHDLIB	=	-lrpc -lnx
ARCHDOBJ	=
ARCHLIB		=	-lrpc -lnx
HASRANLIB	=	f
PVM_ARCH	=	PGON
AR_FRONT	=	ar860
AR_NODE		=	ar860
AR			=	ar860
CC			=	icc
NODECC		=	icc
FORTRAN		=	if77
NODEFORT	=	if77
PVMDEPPATH	=	PGON/
MAKE		=	make
