ARCHCFLAGS	=	-DSYSVSTR -DSYSVSIGNAL -DNOUNIXDOM \
				-DRSHCOMMAND=\"/usr/bin/rcmd\" -DNEEDSFFS \
				-DSYSVBFUNC -DNOWAIT3 -DNOGETDTBLSIZ
ARCHDLIB	=	-lsocket
ARCHDOBJ	=
ARCHLIB		=	-lrpc -lsocket
HASRANLIB	=	
AR			=	ar
PVM_ARCH	=	SCO
MAKE		=	make
