ARCHCFLAGS	=	-D__ST_MT_ERRNO__ -Zsysv-signals -DSYSVSIGNAL \
				-DHASERRORVARS -DNEEDMENDIAN -DHASSTDLIB -DSYSVSTR \
				-DNOGETDTBLSIZ -DNOTMPNAM -DNOWAIT3 -DNOUNIXDOM \
				-DNOSOCKOPT -DNOWAITPID -DRSHCOMMAND=\"rsh\" \
				-DMAXHOSTNAMELEN=256 -DFAKEXDRFLOAT -DFDSETNOTSTRUCT \
				-DSYSERRISCONST
ARCHDLIB	=	-lsocket -lrpc
ARCHDOBJ	=
ARCHLIB		=	-lsocket -lrpc
HASRANLIB	=	f
PVM_ARCH	=	OS2
EXESFX		=	.exe       
CC			=	gcc
MAKE		=	make
