# Warning: Now Subsumed By NETBSDMIPSEL arch.
ARCHCFLAGS	=	-DSOCKADHASLEN -DRSHCOMMAND=\"/usr/bin/rsh\" \
				-DHASERRORVARS -DNOREXEC -DNEEDMENDIAN -DSYSERRISCONST
ARCHDLIB	=
ARCHDOBJ	=
ARCHLIB		=
HASRANLIB	=	t
AR			=	ar
PVM_ARCH	=	NETBSDPMAX
MAKE		=	make
#MAKE		=	gmake
