ARCHCFLAGS	=	-DSOCKADHASLEN -DNOREXEC -DRSHCOMMAND=\"/usr/bin/rsh\" \
				-DHASSTDLIB -DNEEDMENDIAN -DHASERRORVARS \
				-DFAKEXDRFLOAT -DSYSERRISCONST
ARCHDLIB	=
ARCHDOBJ	=
ARCHLIB		=	-lrpcsvc
HASRANLIB	=	t
AR			=	ar
PVM_ARCH	=	FREEBSD
MAKE		=	make
