PVM_ROOT=  C:\PROGRA~1\PVM3.4
PVM_ARCH=  WIN32
CCOMPILER= VISUALC++
FCOMPILER=  DigitalFortran
FLIBDIR=   C:\PROGRA~1\DEVSTU~1\DF\LIB
FINCLUDE=  C:\PROGRA~1\DEVSTU~1\DF\INCLUDE
CLIBDIR=   C:\PROGRA~1\DEVSTU~1\VC\LIB
CINCLUDE=  C:\PROGRA~1\DEVSTU~1\VC\INCLUDE
